#include <stdio.h>
#include <math.h>

  /* Kvadratna jednadzba u obliku x^2 + p*x + q = 0. */

  /* Rjesenja kvadratne jednadzbe obicnim i Vietinim formulama,
     u tipu double, i zaokruzivanjem svih medjurezultata
     (izracunatih u double) na 5 dekadskih znacajnih znamenki. */

  /* Koristi tri funkcije:
      round_to_digits - za zaokruzivanje zadanog broja na
                        digits = 5 dekadskih znacajnih znamenki,
      kvad_double - za rjesavanje kvadratne jednadzbe u double,
      kvad_round - za rjesavanje kvadratne jednadzbe zaokruzivanjem
                   svakog medjurezultata na 5 dekadskih znacajnih znamenki. */


    /* Funckija za zaokruzivanje broja na 5 dekadskih znacajnih znamenki. */

double round_to_digits(double x)
{
    const double base = 10.0;
    const int digits = 5;

        /* Potencija base^(digits) = 10^5 = 100000. */
    const double upp_limit = pow(base, digits);

        /* Potencija base^(digits - 1) = 10^4 = 10000. */
    const double low_limit = pow(base, digits - 1);

    int sign = 1;    /* Predznak = 1 ili -1. */
    int pot = 0;     /* Potencija baze. */
    double x_round;

        /* Test na nulu (u tipu double). */
    if (x == 0.0) return 0.0;

        /* Test negativnog predznaka, dalje x = fabs(x). */
    if (x < 0.0) {
        sign = -1.0;
        x = -x;
    }

        /* Nadji potenciju baze za normalizirani oblik broja,
           tako da je  upp_limit > x >= low_limit. */

    if (x < low_limit) {
        do {
            x *= base;
            ++pot;
        }
        while (x < low_limit);
    }
    else if (x >= upp_limit) {
        do {
            x /= base;
            --pot;
        }
        while (x >= low_limit);
    }

        /* Zaokruzi razlomljeni dio broja. */

    x_round = floor(x);
    if (x - x_round >= 0.5) x_round = x_round + 1.0;    /* = ceil(x). */

        /* Test na nulu (u tipu double). */
    if (x_round == 0.0) return 0.0;

        /* Vrati zaokruzeni broj na polaznu potenciju baze. */
    if (pot > 0) {
        do {
            x_round /= base;
            --pot;
        }
        while (pot > 0);
    }
    else if (pot < 0) {
        do {
            x_round *= base;
            ++pot;
        }
        while (pot < 0);
    }

        /* Vrati zaokruzeni broj s predznakom. */
    if (sign == 1.0)
        return x_round;
    else
        return -x_round;
}


    /* Funckija za rjesavanje kvadratne jednadzbe u double. */

void kvad_double(double p, double q,
                 double *px_1, double *px_2,
                 double *pv_1, double *pv_2)
{
    double diskr, mp_half, x_1, x_2, v_1, v_2, temp;

    mp_half = -p / 2;
    diskr = sqrt(mp_half * mp_half - q);

        /* Obicne formule. */

    x_1 = mp_half + diskr;
    x_2 = mp_half - diskr;

        /* Za usporedbu, poredamo rastuce po apsolutnoj vrijednosti. */
    if (fabs(x_1) > fabs(x_2)) {
        temp = x_1;
        x_1 = x_2;
        x_2 = temp;
    }

        /* Prvo veci po apsolutnoj vrijednosti, drugi iz Vietine formule. */

    if (mp_half >= 0.0)
        v_2 = mp_half + diskr;
    else
        v_2 = mp_half - diskr;
    v_1 = q / v_2;

        /* Vracanje kroz "varijabilne" argumente. */
    *px_1 = x_1;
    *px_2 = x_2;
    *pv_1 = v_1;
    *pv_2 = v_2;
}


    /* Funckija za rjesavanje kvadratne jednadzbe zaokruzivanjem
       svakog medjurezultata na 5 dekadskih znacajnih znamenki. */

void kvad_round(double p, double q,
                double *px_1, double *px_2,
                double *pv_1, double *pv_2)
{
    double diskr, mp_half, x_1, x_2, v_1, v_2, temp;

        /* Zaokruzi koeficijente. */
    p = round_to_digits(p);
    q = round_to_digits(q);

        /* p_half = p / 2; */
    mp_half = round_to_digits(-p / 2);

        /* diskr = sqrt(mp_half * mp_half - q); */

    diskr = round_to_digits(mp_half * mp_half);
    diskr = round_to_digits(diskr - q);
    diskr = round_to_digits(sqrt(diskr));


        /* Obicne formule. */

        /* x_1 = mp_half + diskr; */
    x_1 = round_to_digits(mp_half + diskr);

        /* x_2 = mp_half - diskr; */
    x_2 = round_to_digits(mp_half - diskr);

        /* Za usporedbu, poredamo rastuce po apsolutnoj vrijednosti. */
    if (fabs(x_1) > fabs(x_2)) {
        temp = x_1;
        x_1 = x_2;
        x_2 = temp;
    }

        /* Prvo veci po apsolutnoj vrijednosti, drugi iz Vietine formule. */

    if (mp_half >= 0.0)
            /* v_2 = mp_half + diskr; */
        v_2 = round_to_digits(mp_half + diskr);
    else
            /* v_2 = mp_half - diskr; */
        v_2 = round_to_digits(mp_half - diskr);

        /* v_1 = q / v_2; */
    v_1 = round_to_digits(q / v_2);

        /* Vracanje kroz "varijabilne" argumente. */
    *px_1 = x_1;
    *px_2 = x_2;
    *pv_1 = v_1;
    *pv_2 = v_2;
}


    /* Funckije za ispis koeficijenata i rjesenja. */

void print_coef(double p, double q)
{
    printf("   Egzaktni koeficijenti kvadratne jednadzbe"
           " x^2 + p*x + q = 0\n");
    printf("     p = %8.5f,  q = %8.5f\n", p, q);
    return;
}

void print_roots(double x_1, double x_2)
{
    printf("   Egzaktna rjesenja (iz tipa extended)\n");
    printf("     x_1 = %21.17f\n", x_1);
    printf("     x_2 = %21.17f\n", x_2);
    return;
}

void print_roots_double(double x_1, double x_2,
                        double x_1_exact, double x_2_exact)
{
    double rel_err_1 = 0.0, rel_err_2 = 0.0;

    if (x_1_exact != 0.0)
        rel_err_1 = (x_1_exact - x_1) / x_1_exact;
    if (x_2_exact != 0.0)
        rel_err_2 = (x_2_exact - x_2) / x_2_exact;

    printf("     x_1 = %21.17f,    rel_greska = %13.5e\n", x_1, rel_err_1);
    printf("     x_2 = %21.17f,    rel_greska = %13.5e\n", x_2, rel_err_2);
    return;
}

void print_roots_round(double x_1, double x_2,
                       double x_1_exact, double x_2_exact)
{
    double rel_err_1 = 0.0, rel_err_2 = 0.0;

    if (x_1_exact != 0.0)
        rel_err_1 = (x_1_exact - x_1) / x_1_exact;
    if (x_2_exact != 0.0)
        rel_err_2 = (x_2_exact - x_2) / x_2_exact;

    printf("     x_1 = %10.6f,    rel_greska = %13.5e\n", x_1, rel_err_1);
    printf("     x_2 = %10.6f,    rel_greska = %13.5e\n", x_2, rel_err_2);
    return;
}


    /* Funckija za obradu jedne kvadratne jednadzbe. */

void kvad_jedn(double p, double q,
               double x_1_exact, double x_2_exact)
{
    double x_1, x_2;    /* Izracunata rjesenja obicnim formulama. */
    double v_1, v_2;    /* Izracunata rjesenja Vietinim formulama. */

    print_coef(p, q);
    printf("\n");
    print_roots(x_1_exact, x_2_exact);    /* Egzaktna rjesenja. */

    printf("\n   ***** Double aritmetika\n");
    kvad_double(p, q, &x_1, &x_2, &v_1, &v_2);
    printf("\n   Obicne formule:\n");
    print_roots_double(x_1, x_2, x_1_exact, x_2_exact);
    printf("\n   Vietine formule:\n");
    print_roots_double(v_1, v_2, x_1_exact, x_2_exact);

    printf("\n   ***** Aritmetika s 5 dekadskih znacajnih znamenki\n");
    kvad_round(p, q, &x_1, &x_2, &v_1, &v_2);
    printf("\n   Obicne formule:\n");
    print_roots_round(x_1, x_2, x_1_exact, x_2_exact);
    printf("\n   Vietine formule:\n");
    print_roots_round(v_1, v_2, x_1_exact, x_2_exact);

    printf("\n");
    return;
}


    /* Glavni program. */


int main(void)
{
    double p, q;        /* Koeficijenti kvadratne jednadzbe. */
    double x_1, x_2;    /* Egzaktna rjesenja, izracunata u extended tipu. */


        /* Gautschijev primjer. */

    p = -56.0;
    q = 1.0;
    x_1 =  1.78628407335548633E-0002;
    x_2 =  5.59821371592664451E+0001;

    printf("\n");
    printf(" Gautschijev primjer:\n");
    printf(" ===================\n\n");
    kvad_jedn(p, q, x_1, x_2);


        /* Gautschijev primjer, modifikacija 1. */

    p = -49.0;
    q = 1.0;
    x_1 =  2.04166702127055996E-0002;
    x_2 =  4.89795833297872944E+0001;

    printf("\n");
    printf(" Gautschijev primjer, modifikacija 1:\n");
    printf(" ===================================\n\n");
    kvad_jedn(p, q, x_1, x_2);


        /* Gautschijev primjer, modifikacija 2. */

    p = -490.0;
    q = 1.0;
    x_1 =  2.04082482646116796E-0003;
    x_2 =  4.89997959175173539E+0002;

    printf("\n");
    printf(" Gautschijev primjer, modifikacija 2:\n");
    printf(" ===================================\n\n");
    kvad_jedn(p, q, x_1, x_2);


        /* Skoro dvostruka nultocka u 2. Diskriminanta blizu nule. */
        /* Ovaj primjer NE valja, jer se oduzimanje/kracenje u diskriminanti
           dogadja u double, pa se tek onda zaokruzuje.
           Ovdje su rezultati LOSI, jer racunanje p/2 NIJE egzaktno
                        p/2 = -2.00005,
             izracunati p/2 = -2.0001.
           Kvadriranje jos povecava ovu gresku, pa nakon KRACENJA u
           diskriminanti dobivamoe VELIKU relativnu gresku. */

    p = -4.0001 ;
    q = 4.0 ;
    x_1 =  1.98590777598819763E+0000;
    x_2 =  2.01419222401180237E+0000;

    printf("\n");
    printf(" Skoro dvostruka nultocka u 2:\n");
    printf(" ============================\n\n");
    kvad_jedn(p, q, x_1, x_2);


        /* Skoro dvostruka nultocka u 10. Diskriminanta blizu nule. */
        /* Ovaj primjer NE valja, jer se oduzimanje/kracenje u diskriminanti
           dogadja u double, pa se tek onda zaokruzuje.
           Ovdje su rezultati LOSI, jer racunanje p/2 NIJE egzaktno
                        p/2 = -10.0005,
             izracunati p/2 = -10.001.
           Kvadriranje jos povecava ovu gresku, pa nakon KRACENJA u
           diskriminanti dobivamoe VELIKU relativnu gresku. */

    p = -20.001 ;
    q = 100.0 ;
    x_1 =  9.90049875000781246E+0000;
    x_2 =  1.01005012499921875E+0001;

    printf("\n");
    printf(" Skoro dvostruka nultocka u 10:\n");
    printf(" =============================\n\n");
    kvad_jedn(p, q, x_1, x_2);

    return 0;
}
